#
# All software Copyright 2021 of Techniche Technologies Pty Ltd., All rights reserved.
# Any redistribution or reproduction of part or all of the contents in any form is
# prohibited without express written permission of the company.
#


trap "" 1 2 3 18

if [ -f /usr/local/statseeker/ss/etc/ss.cfg ]; then
   . /usr/local/statseeker/ss/etc/ss.cfg
else
   /usr/local/statseeker/ss/bin/base-log-message -e $0 "Could not load ss.cfg"
   exit 1
fi

export PATH

base_fn_ss_running ()
{
   PID=""

   for lockfile in `find ${SSTMP} -name '*.lock' -print 2> /dev/null`
   do
      . ${lockfile}
      if [ ! -z "${PID}" ]; then
         kill -0 ${PID} 2>/dev/null && return 1
         PID=""
      fi
   done

   return 0
}

base_fn_log_start ()
{
   [ -f ${BASE_TMP}/base-logd.lock ] &&
      base-log-message `basename $0` "started"
}


base_fn_log_end ()
{
   [ -f ${BASE_TMP}/base-logd.lock ] &&
      base-log-message `basename $0` "exited"
}


#
# Daemonise but source the .profile so values are available to scripts
#
# This is potentially more of a problem with perl scripts requiring statseeker
# variables (esp. the PATH variable) when running system commands, instead of
# specifying the full path.
#
# This would also allow us to do things like redirect stdout and stderr to
# logs, or source other environments by default if we want to in the future
#
daemon_as_ss()
{
   local prog=$1

   # NOTE:
   #
   # pass arguments to the sh using "$@" then pass on to prog using "$@"
   #
   # However, sh:
   #  - will gobble the first arg (we get the prog name then rely on sh to gobble it)
   #  - will interpret flags as sh flags and not args (so need to use -- to separate args)

   /usr/sbin/daemon -f -u ${SSUSER} /bin/sh -c ". ${SSROOT}/etc/defaults/dot.profile; exec $prog "'"$@"' -- "$@"
}


#
# Daemonise as root but still source the dot.profile
#
daemon_as_root()
{
   local prog=$1

   # NOTE:
   #
   # pass arguments to the sh using "$@" then pass on to prog using "$@"
   #
   # However, sh:
   #  - will gobble the first arg (we get the prog name then rely on sh to gobble it)
   #  - will interpret flags as sh flags and not args (so need to use -- to separate args)

   /usr/sbin/daemon -f /bin/sh -c ". ${SSROOT}/etc/defaults/dot.profile; exec $prog "'"$@"' -- "$@"
}

ungracefully_kill ()
{
   [ $# -lt 1 ] && {
      base-log-message ungracefully_kill "Script error.  $# parameters supplied"
      return
   }

   PID=$1
   CHILD_PROCESSES=`pgrep -P $PID`
   kill -9 $PID
   # Now kill any children (ungracefully too)
   for child_pid in ${CHILD_PROCESSES}
   do
      # recursive so that children of children are killed.
      ungracefully_kill $child_pid
   done
}

#
# Param 1 - pid to kill
# Param 2 (optional) - Time to wait before kill -9 in seconds
#
# Feel free to use the macros below for adjustment across many files
SS_KILL_FAST=4
SS_KILL_SLOW=15
base_fn_kill_pid ()
{
   [ $# -lt 1 ] && {
      base-log-message base_fn_kill_pid "Script error.  $# parameters supplied"
      return
   }

   PID=$1
   if [ $PID -lt 0 ]; then
      PNAME=`ps -p $((0 - $PID)) -o,comm | sed '/COMM/d'`
   else
      PNAME=`ps -p $PID -o,comm | sed '/COMM/d'`
   fi
   CURTIME=`date '+%s'`
   LOGTIME=$CURTIME
   LOG_INTERVAL=10

   if [ $# -ge 2 ]; then
      ENDTIME=$(($CURTIME + $2))
      if [ $2 -lt 60 ]; then
         LOG_INTERVAL=2
      fi
   else
      ENDTIME=$(($CURTIME + 120))
   fi


   # Send a sigterm to the pid if it exists
   kill -15 $PID  2>/dev/null
   if [ $? -ne 0 ]; then
      return
   fi

   # Wait for the process to exit, logging a message every LOG_INTERVAL seconds.
   while [ "$CURTIME" -lt "$ENDTIME" ]
   do
      kill -15 $PID 2>/dev/null || return

      sleep 0.2

      CURTIME=`date +%s`
      if [ "`expr $CURTIME - $LOGTIME`" -gt $LOG_INTERVAL ]; then
         LOGTIME=$CURTIME
         base-log-message -e base_fn_kill_pid "Waiting for $PNAME ($PID) to exit"
         base-log-message base_fn_kill_pid ""$(ps -p $PID -o,%cpu,state,mwchan,rss,command | sed '/COMM/d')
      fi
   done

   # Kill it ungracefully.
   base-log-message -e base_fn_kill_pid "Process $PNAME ($PID) has not exited.  Killing..."
   ungracefully_kill $PID
}


# Param 1 - Statseeker style lockfile (not system style pidfile)
base_fn_kill_lockfile ()
{
   [ $# -lt 1 ] && {
      base-log-message base_fn_kill_pid "Script error.  $# parameters supplied"
      return
   }
   LOCK_FILE=$1

   if [ -f ${LOCK_FILE} ]; then
      . ${LOCK_FILE}
      if [ -n "$PID" ]; then
         # Use the pid file existence to signal processes to die as well
         rm -f ${LOCK_FILE}
         base_fn_kill_pid $PID $2
      fi
      rm -f ${LOCK_FILE}
   fi
}


[ -z "$SUPPRESS_LOG_START" ] && base_fn_log_start


is_leap_year ()
{
   year=$1

   div_by_four=`expr $year % 4`
   if [ $div_by_four -eq 0 ]
   then
      div_by_hundred=`expr $year % 100`
      if [ $div_by_hundred -eq 0 ]
      then
         div_by_four_hundred=`expr $year % 400`
         if [ $div_by_four_hundred -ne 0 ]
         then
            echo "0"
            return 0
         fi
      fi
      echo "1"
      return 1
   fi
   echo "0"
   return 0
}


days_in_month ()
{
   year=$1
   month=$2

   case $month in
   1|3|5|7|8|10|12)
      echo 31
      return 31
      ;;
   4|6|9|11)
      echo 30
      return 30
      ;;
   2)
      leap_year=`is_leap_year $year`
      result=`expr 28 + $leap_year`
      echo $result
      return $result
      ;;
   *)
      echo 0
      return 0
      ;;
   esac
}


list_ethernet_interfaces ()
{
   for interface in `ifconfig -l`
   do
      ifconfig $interface | grep ether > /dev/null
      if [ $? -eq 0 ]; then
         echo $interface
      fi
   done
}


#
# The steps needed to "reset" the timeseries database, or create it if it
# doesn't exist
#
# This includes:
#  - Creating (if needed)
#  - Checking and fixing corruption or errors (if needed)
#  - Make sure all errors are cleared
#  - Adding the idblocktypes and schedules used in Statseeker
#  - Setting timeseries config based on nim-options.cfg options
#
# There are also some other side effects such as setting
#
# The fixing corruption step can be time consuming. It involves first removing
# all volatile files and fixing them. This will generate a lot of error
# messages for all volatile files not just the ones that had a problem.
#
# @param mode  Either "startup" or "shutdown"
#
#              On startup, a procinfo error is interpreted as showing that the
#              server crashed and didn't shut down clearly (since errors are
#              cleared on shutdown). In this case it is very unlikely that the
#              dcache is current, so it is regenerated from the log files.
#
#              On shutdown, procinfo errors are logged and cleared, which is
#              the normal behaviour for procinfo errors.
#
check_reset_timeseries_database ()
{
   NIM_TS_CMD="${SSROOT}/bin/nim-ts"
   NIM_TS_DIR="${SSHOME}/nim/data/nim-ts"
   RET=0

   MODE=${1:-"shutdown"}

   #
   # Create the timeseries db (if needed)
   #
   if [ ! -d "${NIM_TS_DIR}" ]; then
      su ${SSUSER} -c "${SSROOT}/bin/nim-ts create";
      RET=1
   fi

   #
   # Check that the database can at least open (the config list command is a lot quicker than dbinfo)
   #
   CAN_OPEN=$(su ${SSUSER} -c "${NIM_TS_CMD} config list >/dev/null && echo 1 || echo 0")

   #
   # Check/Fix corruption or errors (if needed)
   #
   if [ ${MODE} == "always" ]; then
      NERRORS=1
      NERRORS_v2=1
   else
      NERRORS=$(su ${SSUSER} -c "${NIM_TS_CMD} validate report files" | grep -v '.z$' | grep -c ERROR)
      NERRORS_v2=0
      if [ "${MODE}" == "startup" ]; then
         NERRORS_v2=$(su ${SSUSER} -c "${NIM_TS_CMD} validate report procinfo" | grep -c ERROR)
      fi
   fi

   if [ $CAN_OPEN -eq 0 ] || [ $NERRORS -gt 0 ] || [ $NERRORS_v2 -gt 0 ]; then

      RET=1

      if [ ${MODE} == "always" ]; then
         base-log-message -e $NAME "INFO: Repairing the timeseries database files"
      elif [ $CAN_OPEN -eq 0 ] || [ $NERRORS -gt 0 ]; then
         base-log-message -e $NAME "ERROR: Corrupt database detected - Repairing"
      else
         base-log-message -e $NAME "ERROR: Possible crash detected - Recreating volatile database files"
      fi

      su ${SSUSER} -c "${NIM_TS_CMD} recover"
   fi

   if [ $IS_HA -eq 1 ]; then
      HOST_NAME_SHORT=$(hostname | sed 's/\..*//g')
      su ${SSUSER} -c "/usr/local/statseeker/ss/cgi/nim-msg 'mode=create' 'type=info' 'message=HA: $HOST_NAME_SHORT becoming Master' >/dev/null"
   fi


   #
   # Make sure all errors are cleared
   #
   su ${SSUSER} -c "${NIM_TS_CMD} validate clear procinfo";
   su ${SSUSER} -c "${NIM_TS_CMD} validate clear lock 0 1";

   # Don't clear files. If something goes wrong, this could clear the whole database.
   # The "recover" should have removed all unneeded files
   # Instead, make it obvious if there are unused files and what command can be manually run to clear it
   #
   # Also, if something has gone wrong, this will spit out 1000's of messages which hopefully won't be missed.
   #
   # TODO: how to not run twice? without creating tmp files
   NERRORS=$(su ${SSUSER} -c "${NIM_TS_CMD} validate report files" | wc -l)
   if [ "$NERRORS" -gt 0 ]; then
      su ${SSUSER} -c "${NIM_TS_CMD} validate report files"
      echo ""
      echo "  NOTE: File errors might be able to be cleared by running"
      echo "    nim-ts validate fix files"
      echo "    nim-ts validate clear files"
      echo ""
      echo " Please shut down Statseeker if running these commands"
      echo ""
      echo ""
   fi


   #
   # Add the idblocktypes and schedules used in Statseeker
   #
   if [ -f "${SSHOME}/nim/etc/nim-ts-schedules.cfg" ]; then
      FILENAME="${SSHOME}/nim/etc/nim-ts-schedules.cfg"
   else
      FILENAME="${SSROOT}/etc/defaults/nim-ts-schedules.cfg"
   fi

   su ${SSUSER} -c "${SSROOT}/bin/nim-ts schedule file '${FILENAME}'"


   #
   # Set default timeseries modes based on nim-options.cfg
   #
   echo                           "Setting timeseries database modes"
   base-log-message `basename $0` "Setting timeseries database modes"

   su ${SSUSER} -c "${SSROOT}/bin/ssperl -e 'use nim_discover; update_timeseries_modes(undef, 1);'"

   return $RET
}

# Delete an sqlite database, along with its assoicated temporary files.
# Supply a list of fully qualified path names to an sqlite db file.
delete_sqlite_db()
{
   for DB in "$@"
   do
      rm -f $DB*
   done
}


# Alias to reduce confusion every time I have to make upgrade changes :)
# Returns 0 for true, and 1 or 2 for false
upgrading_from_version_less_than()
{
   local new_version="$1"

   # Is the version we are upgrading from less than the new version
   base-version -q -f ${SSHOME}/.ssversion $new_version
   return $?
}


# Register or update a process with nim-process-manager
proc_manager_set_process()
{
   local proc_name="$1"
   local proc_cmd="$2"

   [ -z "$proc_name" ] && return

   if [ -n "$proc_cmd" ]; then
      /usr/local/bin/python -c "from ssUtils.CLIUtils import proc_manager_set_process; proc_manager_set_process('$proc_name', '$proc_cmd')"
   else
      /usr/local/bin/python -c "from ssUtils.CLIUtils import proc_manager_remove_process; proc_manager_remove_process('$proc_name')"
   fi
}

ss_state_starting()
{
   echo -n '0' > /var/run/statseeker.run
}

ss_state_running()
{
   echo -n `date "+%s"` > /var/run/statseeker.run
}

ss_state_stopped()
{
   rm -f /var/run/statseeker.run
}

ss_state_stopping()
{
   echo -n '-1' > /var/run/statseeker.run
}

ss_state()
{
   RUNFILE='/var/run/statseeker.run'
   if [ ! -f ${RUNFILE} ]; then
      ss_run_state="stopped"
   else
      STATE=`cat ${RUNFILE}`
      case ${STATE} in
         0)
            ss_run_state="starting"
            ;;
         -1)
            ss_run_state="stopping"
            ;;
         *)
            ss_run_state="running"
            ;;
      esac
   fi
}
